@@ -1,17 +1,12 @@
 # This file is part of Hypothesis, which may be found at
 # https://github.com/HypothesisWorks/hypothesis/
 #
-# Most of this work is copyright (C) 2013-2021 David R. MacIver
-# (david@drmaciver.com), but it contains contributions by others. See
-# CONTRIBUTING.rst for a full list of people who may hold copyright, and
-# consult the git log if you need to determine who owns an individual
-# contribution.
+# Copyright the Hypothesis Authors.
+# Individual contributors are listed in AUTHORS.rst and the git log.
 #
 # This Source Code Form is subject to the terms of the Mozilla Public License,
 # v. 2.0. If a copy of the MPL was not distributed with this file, You can
 # obtain one at https://mozilla.org/MPL/2.0/.
-#
-# END HEADER
 
 """A module controlling settings for Hypothesis to use in falsification.
 
@@ -92,11 +87,11 @@ default_variable = DynamicVariable(None)
 
 
 class settingsMeta(type):
-    def __init__(self, *args, **kwargs):
+    def __init__(cls, *args, **kwargs):
         super().__init__(*args, **kwargs)
 
     @property
-    def default(self):
+    def default(cls):
         v = default_variable.value
         if v is not None:
             return v
@@ -105,10 +100,10 @@ class settingsMeta(type):
             assert default_variable.value is not None
         return default_variable.value
 
-    def _assign_default_internal(self, value):
+    def _assign_default_internal(cls, value):
         default_variable.value = value
 
-    def __setattr__(self, name, value):
+    def __setattr__(cls, name, value):
         if name == "default":
             raise AttributeError(
                 "Cannot assign to the property settings.default - "
@@ -121,7 +116,7 @@ class settingsMeta(type):
                 "settings with settings.load_profile, or use @settings(...) "
                 "to decorate your test instead."
             )
-        return type.__setattr__(self, name, value)
+        return super().__setattr__(name, value)
 
 
 class settings(metaclass=settingsMeta):
@@ -153,19 +148,19 @@ class settings(metaclass=settingsMeta):
         # The intended use is "like **kwargs, but more tractable for tooling".
         max_examples: int = not_set,  # type: ignore
         derandomize: bool = not_set,  # type: ignore
-        database: Union[None, "ExampleDatabase"] = not_set,  # type: ignore
+        database: Optional["ExampleDatabase"] = not_set,  # type: ignore
         verbosity: "Verbosity" = not_set,  # type: ignore
         phases: Collection["Phase"] = not_set,  # type: ignore
         stateful_step_count: int = not_set,  # type: ignore
         report_multiple_bugs: bool = not_set,  # type: ignore
         suppress_health_check: Collection["HealthCheck"] = not_set,  # type: ignore
-        deadline: Union[None, int, float, datetime.timedelta] = not_set,  # type: ignore
+        deadline: Union[int, float, datetime.timedelta, None] = not_set,  # type: ignore
         print_blob: bool = not_set,  # type: ignore
     ) -> None:
         if parent is not None:
             check_type(settings, parent, "parent")
         if derandomize not in (not_set, False):
-            if database not in (not_set, None):
+            if database not in (not_set, None):  # type: ignore
                 raise InvalidArgument(
                     "derandomize=True implies database=None, so passing "
                     f"database={database!r} too is invalid."
@@ -212,7 +207,7 @@ class settings(metaclass=settingsMeta):
                     )
                 setattr(test, attr_name, True)
                 _test.TestCase.settings = self
-                return test
+                return test  # type: ignore
             else:
                 raise InvalidArgument(
                     "@settings(...) can only be used as a decorator on "
@@ -435,12 +430,12 @@ example database implementations, and how to define custom implementations.
 
 @unique
 class Phase(IntEnum):
-    explicit = 0
-    reuse = 1
-    generate = 2
-    target = 3
-    shrink = 4
-    explain = 5
+    explicit = 0  #: controls whether explicit examples are run.
+    reuse = 1  #: controls whether previous examples will be reused.
+    generate = 2  #: controls whether new examples will be generated.
+    target = 3  #: controls whether examples will be mutated for targeting.
+    shrink = 4  #: controls whether examples will be shrunk.
+    explain = 5  #: controls whether Hypothesis attempts to explain test failures.
 
     def __repr__(self):
         return f"Phase.{self.name}"
@@ -491,12 +486,12 @@ class HealthCheck(Enum):
     method defined by :class:`python:unittest.TestCase` (i.e. not a test)."""
 
     function_scoped_fixture = 9
-    """Check if :func:`@given <hypothesis.given>` has been applied to a test
+    """Checks if :func:`@given <hypothesis.given>` has been applied to a test
     with a pytest function-scoped fixture. Function-scoped fixtures run once
     for the whole function, not once per example, and this is usually not what
     you want.
 
-    Because of this limitation, tests that need to need to set up or reset
+    Because of this limitation, tests that need to set up or reset
     state for every example need to do so manually within the test itself,
     typically using an appropriate context manager.
 
