@@ -1,24 +1,19 @@
 # This file is part of Hypothesis, which may be found at
 # https://github.com/HypothesisWorks/hypothesis/
 #
-# Most of this work is copyright (C) 2013-2021 David R. MacIver
-# (david@drmaciver.com), but it contains contributions by others. See
-# CONTRIBUTING.rst for a full list of people who may hold copyright, and
-# consult the git log if you need to determine who owns an individual
-# contribution.
+# Copyright the Hypothesis Authors.
+# Individual contributors are listed in AUTHORS.rst and the git log.
 #
 # This Source Code Form is subject to the terms of the Mozilla Public License,
 # v. 2.0. If a copy of the MPL was not distributed with this file, You can
 # obtain one at https://mozilla.org/MPL/2.0/.
-#
-# END HEADER
 
 import math
-import traceback
 from typing import NoReturn, Union
 
 from hypothesis import Verbosity, settings
-from hypothesis.errors import CleanupFailed, InvalidArgument, UnsatisfiedAssumption
+from hypothesis.errors import InvalidArgument, UnsatisfiedAssumption
+from hypothesis.internal.compat import BaseExceptionGroup
 from hypothesis.internal.conjecture.data import ConjectureData
 from hypothesis.internal.validation import check_type
 from hypothesis.reporting import report, verbose_report
@@ -79,18 +74,16 @@ class BuildContext:
 
     def __exit__(self, exc_type, exc_value, tb):
         self.assign_variable.__exit__(exc_type, exc_value, tb)
-        if self.close() and exc_type is None:
-            raise CleanupFailed()
-
-    def close(self):
-        any_failed = False
+        errors = []
         for task in self.tasks:
             try:
                 task()
-            except BaseException:
-                any_failed = True
-                report(traceback.format_exc())
-        return any_failed
+            except BaseException as err:
+                errors.append(err)
+        if errors:
+            if len(errors) == 1:
+                raise errors[0] from exc_value
+            raise BaseExceptionGroup("Cleanup failed", errors) from exc_value
 
 
 def cleanup(teardown):
@@ -116,7 +109,7 @@ def should_note():
 
 
 def note(value: str) -> None:
-    """Report this value in the final execution."""
+    """Report this value for the minimal failing example."""
     if should_note():
         report(value)
 
@@ -185,8 +178,8 @@ def target(observation: Union[int, float], *, label: str = "") -> Union[int, flo
 
     if label in context.data.target_observations:
         raise InvalidArgument(
-            "Calling target(%r, label=%r) would overwrite target(%r, label=%r)"
-            % (observation, label, context.data.target_observations[label], label)
+            f"Calling target({observation!r}, label={label!r}) would overwrite "
+            f"target({context.data.target_observations[label]!r}, label={label!r})"
         )
     else:
         context.data.target_observations[label] = observation
