@@ -1,17 +1,12 @@
 # This file is part of Hypothesis, which may be found at
 # https://github.com/HypothesisWorks/hypothesis/
 #
-# Most of this work is copyright (C) 2013-2021 David R. MacIver
-# (david@drmaciver.com), but it contains contributions by others. See
-# CONTRIBUTING.rst for a full list of people who may hold copyright, and
-# consult the git log if you need to determine who owns an individual
-# contribution.
+# Copyright the Hypothesis Authors.
+# Individual contributors are listed in AUTHORS.rst and the git log.
 #
 # This Source Code Form is subject to the terms of the Mozilla Public License,
 # v. 2.0. If a copy of the MPL was not distributed with this file, You can
 # obtain one at https://mozilla.org/MPL/2.0/.
-#
-# END HEADER
 
 """
 Writing tests with Hypothesis frees you from the tedium of deciding on and
@@ -47,12 +42,18 @@ generally do their best to write you a useful test.  You can also use
     Options:
       --roundtrip                start by testing write/read or encode/decode!
       --equivalent               very useful when optimising or refactoring code
-      --idempotent
-      --binary-op
+      --errors-equivalent        --equivalent, but also allows consistent errors
+      --idempotent               check that f(x) == f(f(x))
+      --binary-op                associativity, commutativity, identity element
       --style [pytest|unittest]  pytest-style function, or unittest-style method?
       -e, --except OBJ_NAME      dotted name of exception(s) to ignore
       -h, --help                 Show this message and exit.
 
+.. tip::
+
+    Using a light theme?  Hypothesis respects `NO_COLOR <https://no-color.org/>`__
+    and :envvar:`DJANGO_COLORS=light <django:DJANGO_COLORS>`.
+
 .. note::
 
     The ghostwriter requires :pypi:`black`, but the generated code only
@@ -78,9 +79,11 @@ import sys
 import types
 from collections import OrderedDict, defaultdict
 from itertools import permutations, zip_longest
+from keyword import iskeyword
 from string import ascii_lowercase
 from textwrap import dedent, indent
 from typing import (
+    TYPE_CHECKING,
     Any,
     Callable,
     Dict,
@@ -99,8 +102,10 @@ import black
 from hypothesis import Verbosity, find, settings, strategies as st
 from hypothesis.errors import InvalidArgument
 from hypothesis.internal.compat import get_type_hints
-from hypothesis.internal.reflection import is_mock
+from hypothesis.internal.reflection import get_signature, is_mock
 from hypothesis.internal.validation import check_type
+from hypothesis.provisional import domains
+from hypothesis.strategies._internal.collections import ListStrategy
 from hypothesis.strategies._internal.core import BuildsStrategy
 from hypothesis.strategies._internal.flatmapped import FlatMapStrategy
 from hypothesis.strategies._internal.lazy import LazyStrategy, unwrap_strategies
@@ -111,7 +116,15 @@ from hypothesis.strategies._internal.strategies import (
     SampledFromStrategy,
 )
 from hypothesis.strategies._internal.types import _global_type_lookup, is_generic_type
-from hypothesis.utils.conventions import InferType, infer
+from hypothesis.utils.conventions import infer
+
+if sys.version_info >= (3, 10):  # pragma: no cover
+    from types import EllipsisType as InferType
+elif TYPE_CHECKING:
+    from builtins import ellipsis as InferType
+else:
+    InferType = type(Ellipsis)
+
 
 IMPORT_SECTION = """
 # This test code was written by the `hypothesis.extra.ghostwriter` module
@@ -126,14 +139,15 @@ def test_{test_kind}_{func_name}({arg_names}):
 {test_body}
 """
 
-SUPPRESS_BLOCK = """\
+SUPPRESS_BLOCK = """
 try:
 {test_body}
 except {exceptions}:
     reject()
-""".rstrip()
+""".strip()
 
 Except = Union[Type[Exception], Tuple[Type[Exception], ...]]
+ImportSet = Set[Union[str, Tuple[str, str]]]
 RE_TYPES = (type(re.compile(".")), type(re.match(".", "abc")))
 _quietly_settings = settings(
     database=None,
@@ -171,6 +185,22 @@ def _check_except(except_: Except) -> Tuple[Type[Exception], ...]:
     return (except_,)
 
 
+def _exception_string(except_: Tuple[Type[Exception], ...]) -> Tuple[ImportSet, str]:
+    if not except_:
+        return set(), ""
+    exceptions = []
+    imports: ImportSet = set()
+    for ex in _dedupe_exceptions(except_):
+        if ex.__qualname__ in dir(builtins):
+            exceptions.append(ex.__qualname__)
+        else:
+            imports.add(ex.__module__)
+            exceptions.append(_get_qualname(ex, include_module=True))
+    return imports, (
+        "(" + ", ".join(exceptions) + ")" if len(exceptions) > 1 else exceptions[0]
+    )
+
+
 def _check_style(style: str) -> None:
     if style not in ("pytest", "unittest"):
         raise InvalidArgument(f"Valid styles are 'pytest' or 'unittest', got {style!r}")
@@ -194,34 +224,15 @@ def _exceptions_from_docstring(doc: str) -> Tuple[Type[Exception], ...]:
     return tuple(_dedupe_exceptions(tuple(raises)))
 
 
-# Simple strategies to guess for common argument names - we wouldn't do this in
-# builds() where strict correctness is required, but we only use these guesses
-# when the alternative is nothing() to force user edits anyway.
-#
-# This table was constructed manually after skimming through the documentation
-# for the builtins and a few stdlib modules.  Future enhancements could be based
-# on analysis of type-annotated code to detect arguments which almost always
-# take values of a particular type.
-_GUESS_STRATEGIES_BY_NAME = (
-    (st.integers(0, 32), ["ndims"]),
-    (st.booleans(), ["keepdims"]),
-    (st.text(), ["name", "filename", "fname"]),
-    (st.floats(), ["real", "imag"]),
-    (st.functions(), ["function", "func", "f"]),
-    (st.functions(returns=st.booleans(), pure=True), ["pred", "predicate"]),
-    (st.iterables(st.integers()) | st.iterables(st.text()), ["iterable"]),
-)
-
-
 def _type_from_doc_fragment(token: str) -> Optional[type]:
     # Special cases for "integer" and for numpy array-like and dtype
     if token == "integer":
         return int
     if "numpy" in sys.modules:
         if re.fullmatch(r"[Aa]rray[-_ ]?like", token):
-            return sys.modules["numpy"].ndarray  # type: ignore
+            return sys.modules["numpy"].ndarray
         elif token == "dtype":
-            return sys.modules["numpy"].dtype  # type: ignore
+            return sys.modules["numpy"].dtype
     # Natural-language syntax, e.g. "sequence of integers"
     coll_match = re.fullmatch(r"(\w+) of (\w+)", token)
     if coll_match is not None:
@@ -250,9 +261,9 @@ def _strategy_for(
     # - `b (list, tuple, or None): ...`
     # - `c : {"foo", "bar", or None}`
     for pattern in (
-        fr"^\s*\:type\s+{param.name}\:\s+(.+)",  # RST-style
-        fr"^\s*{param.name} \((.+)\):",  # Google-style
-        fr"^\s*{param.name} \: (.+)",  # Numpy-style
+        rf"^\s*\:type\s+{param.name}\:\s+(.+)",  # RST-style
+        rf"^\s*{param.name} \((.+)\):",  # Google-style
+        rf"^\s*{param.name} \: (.+)",  # Numpy-style
     ):
         match = re.search(pattern, docstring, flags=re.MULTILINE)
         if match is None:
@@ -306,14 +317,124 @@ def _strategy_for(
         # failures in cases like the `flags` argument to regex functions.
         # Better in to keep it simple, and let the user elaborate if desired.
         return st.just(param.default)
-    # If there's no annotation and no default value, we check against a table
-    # of guesses of simple strategies for common argument names.
-    if "string" in param.name and "as" not in param.name:
+    return _guess_strategy_by_argname(name=param.name.lower())
+
+
+# fmt: off
+BOOL_NAMES = (
+    "keepdims", "verbose", "debug", "force", "train", "training", "trainable", "bias",
+    "shuffle", "show", "load", "pretrained", "save", "overwrite", "normalize",
+    "reverse", "success", "enabled", "strict", "copy", "quiet", "required", "inplace",
+    "recursive", "enable", "active", "create", "validate", "refresh", "use_bias",
+)
+POSITIVE_INTEGER_NAMES = (
+    "width", "size", "length", "limit", "idx", "stride", "epoch", "epochs", "depth",
+    "pid", "steps", "iteration", "iterations", "vocab_size", "ttl", "count",
+)
+FLOAT_NAMES = (
+    "real", "imag", "alpha", "theta", "beta", "sigma", "gamma", "angle", "reward",
+    "tau", "temperature",
+)
+STRING_NAMES = (
+    "text", "txt", "password", "label", "prefix", "suffix", "desc", "description",
+    "str", "pattern", "subject", "reason", "comment", "prompt", "sentence", "sep",
+)
+# fmt: on
+
+
+def _guess_strategy_by_argname(name: str) -> st.SearchStrategy:
+    """
+    If all else fails, we try guessing a strategy based on common argument names.
+
+    We wouldn't do this in builds() where strict correctness is required, but for
+    the ghostwriter we accept "good guesses" since the user would otherwise have
+    to change the strategy anyway - from `nothing()` - if we refused to guess.
+
+    A "good guess" is _usually correct_, and _a reasonable mistake_ if not.
+    The logic below is therefore based on a manual reading of the builtins and
+    some standard-library docs, plus the analysis of about three hundred million
+    arguments in https://github.com/HypothesisWorks/hypothesis/issues/3311
+    """
+    # Special-cased names
+    if name in ("function", "func", "f"):
+        return st.functions()
+    if name in ("pred", "predicate"):
+        return st.functions(returns=st.booleans(), pure=True)
+    if name in ("iterable",):
+        return st.iterables(st.integers()) | st.iterables(st.text())
+    if name in ("list", "lst", "ls"):
+        return st.lists(st.nothing())
+    if name in ("object",):
+        return st.builds(object)
+    if "uuid" in name:
+        return st.uuids().map(str)
+
+    # Names which imply the value is a boolean
+    if name.startswith("is_") or name in BOOL_NAMES:
+        return st.booleans()
+
+    # Names which imply that the value is a number, perhaps in a particular range
+    if name in ("amount", "threshold", "number", "num"):
+        return st.integers() | st.floats()
+
+    if name in ("port",):
+        return st.integers(0, 2**16 - 1)
+    if (
+        name.endswith("_size")
+        or (name.endswith("size") and "_" not in name)
+        or re.fullmatch(r"n(um)?_[a-z_]*s", name)
+        or name in POSITIVE_INTEGER_NAMES
+    ):
+        return st.integers(min_value=0)
+    if name in ("offset", "seed", "dim", "total", "priority"):
+        return st.integers()
+
+    if name in ("learning_rate", "dropout", "dropout_rate", "epsilon", "eps", "prob"):
+        return st.floats(0, 1)
+    if name in ("lat", "latitude"):
+        return st.floats(-90, 90)
+    if name in ("lon", "longitude"):
+        return st.floats(-180, 180)
+    if name in ("radius", "tol", "tolerance", "rate"):
+        return st.floats(min_value=0)
+    if name in FLOAT_NAMES:
+        return st.floats()
+
+    # Names which imply that the value is a string
+    if name in ("host", "hostname"):
+        return domains()
+    if name in ("email",):
+        return st.emails()
+    if name in ("word", "slug", "api_key"):
+        return st.from_regex(r"\w+", fullmatch=True)
+    if name in ("char", "character"):
+        return st.characters()
+
+    if (
+        "file" in name
+        or "path" in name
+        or name.endswith("_dir")
+        or name in ("fname", "dir", "dirname", "directory", "folder")
+    ):
+        # Common names for filesystem paths: these are usually strings, but we
+        # don't want to make strings more convenient than pathlib.Path.
+        return st.nothing()
+
+    if (
+        name.endswith("_name")
+        or (name.endswith("name") and "_" not in name)
+        or ("string" in name and "as" not in name)
+        or name.endswith("label")
+        or name in STRING_NAMES
+    ):
         return st.text()
-    for strategy, names in _GUESS_STRATEGIES_BY_NAME:
-        if param.name in names:
-            assert isinstance(strategy, st.SearchStrategy)
-            return strategy
+
+    # Last clever idea: maybe we're looking a plural, and know the singular:
+    if re.fullmatch(r"\w*[^s]s", name):
+        elems = _guess_strategy_by_argname(name[:-1])
+        if not elems.is_empty:
+            return st.lists(elems)
+
     # And if all that failed, we'll return nothing() - the user will have to
     # fill this in by hand, and we'll leave a comment to that effect later.
     return st.nothing()
@@ -323,7 +444,7 @@ def _get_params(func: Callable) -> Dict[str, inspect.Parameter]:
     """Get non-vararg parameters of `func` as an ordered dict."""
     var_param_kinds = (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD)
     try:
-        params = list(inspect.signature(func).parameters.values())
+        params = list(get_signature(func).parameters.values())
     except Exception:
         if (
             isinstance(func, (types.BuiltinFunctionType, types.BuiltinMethodType))
@@ -331,18 +452,29 @@ def _get_params(func: Callable) -> Dict[str, inspect.Parameter]:
             and isinstance(func.__doc__, str)
         ):
             # inspect.signature doesn't work on all builtin functions or methods.
-            # In such cases, including the operator module on Python 3.6, we can try
-            # to reconstruct simple signatures from the docstring.
-            pattern = rf"^{func.__name__}\(([a-z]+(, [a-z]+)*)(, \\)?\)"
-            args = re.match(pattern, func.__doc__)
-            if args is None:
+            # In such cases, we can try to reconstruct simple signatures from the docstring.
+            match = re.match(rf"^{func.__name__}\((.+?)\)", func.__doc__)
+            if match is None:
                 raise
-            params = [
-                # Note that we assume that the args are positional-only regardless of
-                # whether the signature shows a `/`, because this is often the case.
-                inspect.Parameter(name=name, kind=inspect.Parameter.POSITIONAL_ONLY)
-                for name in args.group(1).split(", ")
-            ]
+            args = match.group(1).replace("[", "").replace("]", "")
+            params = []
+            # Even if the signature doesn't contain a /, we assume that arguments
+            # are positional-only until shown otherwise - the / is often omitted.
+            kind: inspect._ParameterKind = inspect.Parameter.POSITIONAL_ONLY
+            for arg in args.split(", "):
+                arg, *_ = arg.partition("=")
+                arg = arg.strip()
+                if arg == "/":
+                    kind = inspect.Parameter.POSITIONAL_OR_KEYWORD
+                    continue
+                if arg.startswith("*") or arg == "...":
+                    kind = inspect.Parameter.KEYWORD_ONLY
+                    continue  # we omit *varargs, if there are any
+                if iskeyword(arg.lstrip("*")) or not arg.lstrip("*").isidentifier():
+                    print(repr(args))
+                    break  # skip all subsequent params if this name is invalid
+                params.append(inspect.Parameter(name=arg, kind=kind))
+
         elif _is_probably_ufunc(func):
             # `inspect.signature` doesn't work on ufunc objects, but we can work out
             # what the required parameters would look like if it did.
@@ -449,9 +581,6 @@ def _imports_for_object(obj):
         name = _get_qualname(obj).split(".")[0]
         return {(_get_module(obj), name)}
     except Exception:
-        with contextlib.suppress(AttributeError):
-            if obj.__module__ == "typing":  # only on CPython 3.6
-                return {("typing", getattr(obj, "__name__", obj.name))}
         return set()
 
 
@@ -463,6 +592,8 @@ def _imports_for_strategy(strategy):
             return _imports_for_object(strategy._LazyStrategy__args[0])
         elif _get_module(strategy.function).startswith("hypothesis.extra."):
             return {(_get_module(strategy.function), strategy.function.__name__)}
+            module = _get_module(strategy.function).replace("._array_helpers", ".numpy")
+            return {(module, strategy.function.__name__)}
 
     imports = set()
     strategy = unwrap_strategies(strategy)
@@ -496,6 +627,9 @@ def _imports_for_strategy(strategy):
         for obj in strategy.elements:
             imports |= _imports_for_object(obj)
 
+    if isinstance(strategy, ListStrategy):
+        imports |= _imports_for_strategy(strategy.element_strategy)
+
     return imports
 
 
@@ -539,11 +673,25 @@ def _valid_syntax_repr(strategy):
 KNOWN_FUNCTION_LOCATIONS: Dict[object, str] = {}
 
 
+def _get_module_helper(obj):
+    # Get the __module__ attribute of the object, and return the first ancestor module
+    # which contains the object; falling back to the literal __module__ if none do.
+    # The goal is to show location from which obj should usually be accessed, rather
+    # than what we assume is an internal submodule which defined it.
+    module_name = obj.__module__
+    dots = [i for i, c in enumerate(module_name) if c == "."] + [None]
+    for idx in dots:
+        if getattr(sys.modules.get(module_name[:idx]), obj.__name__, None) is obj:
+            KNOWN_FUNCTION_LOCATIONS[obj] = module_name[:idx]
+            return module_name[:idx]
+    return module_name
+
+
 def _get_module(obj):
     if obj in KNOWN_FUNCTION_LOCATIONS:
         return KNOWN_FUNCTION_LOCATIONS[obj]
     try:
-        return obj.__module__
+        return _get_module_helper(obj)
     except AttributeError:
         if not _is_probably_ufunc(obj):
             raise
@@ -615,11 +763,12 @@ def _make_test_body(
     assertions: str = "",
     style: str,
     given_strategies: Optional[Mapping[str, Union[str, st.SearchStrategy]]] = None,
-) -> Tuple[Set[Union[str, Tuple[str, str]]], str]:
+    imports: Optional[ImportSet] = None,
+) -> Tuple[ImportSet, str]:
     # A set of modules to import - we might add to this later.  The import code
     # is written later, so we can have one import section for multiple magic()
     # test functions.
-    imports = {_get_module(f) for f in funcs}
+    imports = (imports or set()) | {_get_module(f) for f in funcs}
 
     # Get strategies for all the arguments to each function we're testing.
     with _with_any_registered():
@@ -633,20 +782,13 @@ def _make_test_body(
 
     if except_:
         # Convert to strings, either builtin names or qualified names.
-        exceptions = []
-        for ex in _dedupe_exceptions(except_):
-            if ex.__qualname__ in dir(builtins):
-                exceptions.append(ex.__qualname__)
-            else:
-                imports.add(ex.__module__)
-                exceptions.append(_get_qualname(ex, include_module=True))
+        imp, exc_string = _exception_string(except_)
+        imports.update(imp)
         # And finally indent the existing test body into a try-except block
         # which catches these exceptions and calls `hypothesis.reject()`.
         test_body = SUPPRESS_BLOCK.format(
             test_body=indent(test_body, prefix="    "),
-            exceptions="(" + ", ".join(exceptions) + ")"
-            if len(exceptions) > 1
-            else exceptions[0],
+            exceptions=exc_string,
         )
 
     if assertions:
@@ -674,7 +816,7 @@ def _make_test_body(
     return imports, body
 
 
-def _make_test(imports: Set[Union[str, Tuple[str, str]]], body: str) -> str:
+def _make_test(imports: ImportSet, body: str) -> str:
     # Discarding "builtins." and "__main__" probably isn't particularly useful
     # for user code, but important for making a good impression in demos.
     body = body.replace("builtins.", "").replace("__main__.", "")
@@ -688,7 +830,8 @@ def _make_test(imports: Set[Union[str, Tuple[str, str]]], body: str) -> str:
     direct = {f"import {i}" for i in imports - do_not_import if isinstance(i, str)}
     from_imports = defaultdict(set)
     for module, name in {i for i in imports if isinstance(i, tuple)}:
-        from_imports[module].add(name)
+        if not (module.startswith("hypothesis.strategies") and name in st.__all__):
+            from_imports[module].add(name)
     from_ = {
         "from {} import {}".format(module, ", ".join(sorted(names)))
         for module, names in from_imports.items()
@@ -764,10 +907,15 @@ def magic(
     for thing in modules_or_functions:
         if callable(thing):
             functions.add(thing)
+            # class need to be added for exploration
+            if inspect.isclass(thing):
+                funcs: List[Optional[Any]] = [thing]
+            else:
+                funcs = []
         elif isinstance(thing, types.ModuleType):
             if hasattr(thing, "__all__"):
-                funcs = [getattr(thing, name, None) for name in thing.__all__]  # type: ignore
-            else:
+                funcs = [getattr(thing, name, None) for name in thing.__all__]
+            elif hasattr(thing, "__package__"):
                 pkg = thing.__package__
                 funcs = [
                     v
@@ -779,22 +927,35 @@ def magic(
                 ]
                 if pkg and any(getattr(f, "__module__", pkg) == pkg for f in funcs):
                     funcs = [f for f in funcs if getattr(f, "__module__", pkg) == pkg]
-            for f in funcs:
-                try:
-                    if (
-                        (not is_mock(f))
-                        and callable(f)
-                        and _get_params(f)
-                        and not isinstance(f, enum.EnumMeta)
-                    ):
-                        functions.add(f)
-                        if getattr(thing, "__name__", None):
-                            KNOWN_FUNCTION_LOCATIONS[f] = thing.__name__
-                except (TypeError, ValueError):
-                    pass
         else:
             raise InvalidArgument(f"Can't test non-module non-callable {thing!r}")
 
+        for f in list(funcs):
+            if inspect.isclass(f):
+                funcs += [
+                    v.__get__(f)
+                    for k, v in vars(f).items()
+                    if hasattr(v, "__func__")
+                    and not is_mock(v)
+                    and not k.startswith("_")
+                ]
+        for f in funcs:
+            try:
+                if (
+                    (not is_mock(f))
+                    and callable(f)
+                    and _get_params(f)
+                    and not isinstance(f, enum.EnumMeta)
+                ):
+                    functions.add(f)
+                    if getattr(thing, "__name__", None):
+                        if inspect.isclass(thing):
+                            KNOWN_FUNCTION_LOCATIONS[f] = _get_module_helper(thing)
+                        else:
+                            KNOWN_FUNCTION_LOCATIONS[f] = thing.__name__
+            except (TypeError, ValueError):
+                pass
+
     imports = set()
     parts = []
 
@@ -1058,11 +1219,72 @@ def _make_equiv_body(funcs, except_, style):
     )
 
 
-def equivalent(*funcs: Callable, except_: Except = (), style: str = "pytest") -> str:
+EQUIV_FIRST_BLOCK = """
+try:
+{}
+    exc_type = None
+    target(1, label="input was valid")
+{}except Exception as exc:
+    exc_type = type(exc)
+""".strip()
+
+EQUIV_CHECK_BLOCK = """
+if exc_type:
+    with {ctx}(exc_type):
+{check_raises}
+else:
+{call}
+{compare}
+""".rstrip()
+
+
+def _make_equiv_errors_body(funcs, except_, style):
+    var_names = [f"result_{f.__name__}" for f in funcs]
+    if len(set(var_names)) < len(var_names):
+        var_names = [f"result_{i}_{ f.__name__}" for i, f in enumerate(funcs)]
+
+    first, *rest = funcs
+    first_call = _write_call(first, assign=var_names[0], except_=except_)
+    extra_imports, suppress = _exception_string(except_)
+    extra_imports.add(("hypothesis", "target"))
+    catch = f"except {suppress}:\n    reject()\n" if suppress else ""
+    test_lines = [EQUIV_FIRST_BLOCK.format(indent(first_call, prefix="    "), catch)]
+
+    for vname, f in zip(var_names[1:], rest):
+        if style == "pytest":
+            ctx = "pytest.raises"
+            extra_imports.add("pytest")
+        else:
+            assert style == "unittest"
+            ctx = "self.assertRaises"
+        block = EQUIV_CHECK_BLOCK.format(
+            ctx=ctx,
+            check_raises=indent(_write_call(f, except_=()), "        "),
+            call=indent(_write_call(f, assign=vname, except_=()), "    "),
+            compare=indent(_assert_eq(style, var_names[0], vname), "    "),
+        )
+        test_lines.append(block)
+
+    imports, source_code = _make_test_body(
+        *funcs,
+        test_body="\n".join(test_lines),
+        except_=(),
+        ghost="equivalent",
+        style=style,
+    )
+    return imports | extra_imports, source_code
+
+
+def equivalent(
+    *funcs: Callable,
+    allow_same_errors: bool = False,
+    except_: Except = (),
+    style: str = "pytest",
+) -> str:
     """Write source code for a property-based test of ``funcs``.
 
-    The resulting test checks that calling each of the functions gives
-    the same result.  This can be used as a classic 'oracle', such as testing
+    The resulting test checks that calling each of the functions returns
+    an equal value.  This can be used as a classic 'oracle', such as testing
     a fast sorting algorithm against the :func:`python:sorted` builtin, or
     for differential testing where none of the compared functions are fully
     trusted but any difference indicates a bug (e.g. running a function on
@@ -1071,15 +1293,25 @@ def equivalent(*funcs: Callable, except_: Except = (), style: str = "pytest") ->
     The functions should have reasonably similar signatures, as only the
     common parameters will be passed the same arguments - any other parameters
     will be allowed to vary.
+
+    If allow_same_errors is True, then the test will pass if calling each of
+    the functions returns an equal value, *or* if the first function raises an
+    exception and each of the others raises an exception of the same type.
+    This relaxed mode can be useful for code synthesis projects.
     """
     if len(funcs) < 2:
         raise InvalidArgument("Need at least two functions to compare.")
     for i, f in enumerate(funcs):
         if not callable(f):
             raise InvalidArgument(f"Got non-callable funcs[{i}]={f!r}")
+    check_type(bool, allow_same_errors, "allow_same_errors")
     except_ = _check_except(except_)
     _check_style(style)
-    return _make_test(*_make_equiv_body(funcs, except_, style))
+    if allow_same_errors and not any(issubclass(Exception, ex) for ex in except_):
+        imports, source_code = _make_equiv_errors_body(funcs, except_, style)
+    else:
+        imports, source_code = _make_equiv_body(funcs, except_, style)
+    return _make_test(imports, source_code)
 
 
 X = TypeVar("X")
@@ -1156,7 +1388,7 @@ def _make_binop_body(
     distributes_over: Optional[Callable[[X, X], X]] = None,
     except_: Tuple[Type[Exception], ...],
     style: str,
-) -> Tuple[Set[Union[str, Tuple[str, str]]], str]:
+) -> Tuple[ImportSet, str]:
     strategies = _get_strategies(func)
     operands, b = (strategies.pop(p) for p in list(_get_params(func))[:2])
     if repr(operands) != repr(b):
@@ -1279,10 +1511,10 @@ def _make_binop_body(
 
 
 def ufunc(func: Callable, *, except_: Except = (), style: str = "pytest") -> str:
-    """Write a property-based test for the :np-ref:`array unfunc <ufuncs.html>` ``func``.
+    """Write a property-based test for the :doc:`array ufunc <numpy:reference/ufuncs>` ``func``.
 
-    The resulting test checks that your ufunc or :np-ref:`gufunc
-    <c-api/generalized-ufuncs.html>` has the expected broadcasting and dtype casting
+    The resulting test checks that your ufunc or :doc:`gufunc
+    <numpy:reference/c-api/generalized-ufuncs>` has the expected broadcasting and dtype casting
     behaviour.  You will probably want to add extra assertions, but as with the other
     ghostwriters this gives you a great place to start.
 
@@ -1305,13 +1537,17 @@ def _make_ufunc_body(func, *, except_, style):
         shapes = npst.mutually_broadcastable_shapes(num_shapes=func.nin)
     else:
         shapes = npst.mutually_broadcastable_shapes(signature=func.signature)
+    shapes.function.__module__ = npst.__name__
 
     body = """
     input_shapes, expected_shape = shapes
     input_dtypes, expected_dtype = types.split("->")
-    array_st = [npst.arrays(d, s) for d, s in zip(input_dtypes, input_shapes)]
+    array_strats = [
+        arrays(dtype=dtp, shape=shp, elements={{"allow_nan": True}})
+        for dtp, shp in zip(input_dtypes, input_shapes)
+    ]
 
-    {array_names} = data.draw(st.tuples(*array_st))
+    {array_names} = data.draw(st.tuples(*array_strats))
     result = {call}
     """.format(
         array_names=", ".join(ascii_lowercase[: func.nin]),
@@ -1337,4 +1573,5 @@ def _make_ufunc_body(func, *, except_, style):
         ghost="ufunc" if func.signature is None else "gufunc",
         style=style,
         given_strategies={"data": st.data(), "shapes": shapes, "types": types},
+        imports={("hypothesis.extra.numpy", "arrays")},
     )
